/* This file is part of Mailfromd.
   Copyright (C) 2005, 2006, 2007, 2008, 2009, 2010 Sergey Poznyakoff

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <mailutils/mailutils.h>
#include "libmf.h"
#include "mfdb.h"
#include "mf-dbm.h"
#include "filenames.h"

time_t negative_expire_interval;//FIXME
time_t tempfail_expire_interval = 37*60;/* FIXME: Equals to the sum of the
					   ``hard'' timeouts */
	
static void cache_print_item(const char *email, size_t size,
			     const void *content);
static int cache_expire_item(const void *content);

static struct db_format cache_format_struct = {
	"cache",
	DEFAULT_DATABASE,
	1,
	0,
	DEFAULT_EXPIRE_INTERVAL,
	cache_print_item,
	cache_expire_item
};

struct db_format *cache_format = &cache_format_struct;


struct cache_result {
	time_t timestamp;
	mf_status status;
};

#define EXPIRE_INTERVAL(s)					\
	((s) == mf_success ? cache_format->expire_interval :	\
	 (((s) == mf_temp_failure || (s) == mf_failure)		\
	  ? tempfail_expire_interval : negative_expire_interval))

static char *
format_timestr(time_t timestamp, char *timebuf, size_t bufsize)
{
	struct tm tm;
	gmtime_r(&timestamp, &tm);
	strftime(timebuf, bufsize, "%c", &tm);
	return timebuf;
}

mf_status
cache_get(const char *email)
{
	mf_status rc;
	MF_DBM_FILE db;
	MF_DBM_DATUM key;
	MF_DBM_DATUM contents;
	int readonly = 0;
	int res;

	if (!cache_format->enabled)
		return mf_failure;
	
	mu_debug(cache_format->debug_handle, MU_DEBUG_TRACE5,
		 ("getting cache info for %s", email));
	if (mf_dbm_open(cache_format->dbname, &db, MU_STREAM_RDWR, 0600,
			&readonly)) {
		mu_error(_("mf_dbm_open(%s) failed: %s"), cache_format->dbname,
			 mf_dbm_strerror());
		return mf_failure;
	}
	if (readonly) 
		mu_debug(cache_format->debug_handle, MU_DEBUG_TRACE0, 
                      ("cannot lock %s: switching to read-only mode",
		       cache_format->dbname));
	
	memset (&key, 0, sizeof key);
	memset (&contents, 0, sizeof contents);
	MF_DATUM_PTR (key) = (void*) email;
	MF_DATUM_SIZE (key) = strlen (email) + 1;
	if ((res = mf_dbm_fetch(&db, key, &contents)) == 0) {
		char timebuf[80];
		struct cache_result *res = (struct cache_result *)
			                     MF_DATUM_PTR(contents);
		time_t t = time(NULL);
		mu_debug(cache_format->debug_handle, MU_DEBUG_TRACE5, 
			 ("found status: %s (%d), time: %s",
			  mf_status_str(res->status), res->status,
			  format_timestr(res->timestamp,
					 timebuf, sizeof timebuf)));
		
		if (t - res->timestamp > EXPIRE_INTERVAL(res->status)) {
			if (!readonly) {
				mu_debug(cache_format->debug_handle,
					 MU_DEBUG_TRACE5,
					 ("removing expired entry for %s",
					  email));
				if (mf_dbm_delete(&db, key))
					mu_error(_("cannot remove record `%s' from `%s': %s"),
						 email, db.name, 
		                                 mf_dbm_strerror());
			}
			rc = mf_failure;
		} else {
			rc = res->status;
		}
	} else {
		if (res != MU_ERR_NOENT)
			mu_error(_("cannot fetch record `%s' from `%s': %s"),
				 email, db.name, mf_dbm_strerror());
		rc = mf_failure;
	}
	mf_dbm_datum_free(&contents);
	mf_dbm_close(&db);
	return rc;
}

void
cache_insert(const char *email, mf_status rc)
{
	MF_DBM_FILE db;
	MF_DBM_DATUM key;
	MF_DBM_DATUM contents;
	struct cache_result res;
	char timebuf[80];
	
	if (!cache_format->enabled)
		return;

	time(&res.timestamp);
	res.status = rc;
	mu_debug(cache_format->debug_handle, MU_DEBUG_TRACE5,
		 ("inserting cache info for %s. status=%s (%d), time=%s",
		  email, mf_status_str(rc), rc,
		  format_timestr(res.timestamp, timebuf, sizeof timebuf)));
	if (mf_dbm_open(cache_format->dbname, &db, MU_STREAM_RDWR, 
                        0600, NULL)) {
		mu_error(_("mf_dbm_open(%s) failed: %s"), cache_format->dbname,
			 mf_dbm_strerror());
		return;
	}

	memset(&key, 0, sizeof key);
	memset(&contents, 0, sizeof contents);
	MF_DATUM_PTR(key) = (void*) email;
	MF_DATUM_SIZE(key) = strlen (email) + 1;
	MF_DATUM_PTR(contents) = (void*)&res;
	MF_DATUM_SIZE(contents) = sizeof(res);
	if (mf_dbm_insert(&db, key, contents, 1))
		mu_error(_("cannot insert datum `%s' into `%s': %s"), 
                         email, db.name, mf_dbm_strerror());
	
	mf_dbm_close(&db);
}

static void
cache_print_item(const char *email, size_t size, const void *content)
{
	const struct cache_result *res = content;

	size--; /* Size includes the trailing nul */
	printf("%*.*s", size, size, email);
	printf(" %10.10s ", mf_status_str(res->status));
	format_time_str(stdout, res->timestamp);
	putchar('\n');
}

static int
cache_expire_item(const void *content)
{
	const struct cache_result *res = (struct cache_result *)content;
	return !res || time(NULL) - res->timestamp > EXPIRE_INTERVAL(res->status);
}

mf_status
cache_get2(const char *email, const char *client_addr)
{
	mf_status rc = mf_failure;
	size_t size;
	char *key;
	
	if (!cache_format->enabled)
		return mf_failure;
	size = strlen(email) + 1 + strlen(client_addr) + 1;
	key = malloc(size);
	if (key) {
		strcat(strcat(strcpy(key, email), ":"), client_addr);
		rc = cache_get(key);
		free(key);
	}
	return rc;
}

void
cache_insert2(const char *email, const char *client_addr, mf_status rc)
{
	if (!cache_format->enabled)
		return;
	else { 
		size_t size = strlen(email) + 1 + strlen(client_addr) + 1;
		char *key = malloc(size);
		if (key) {
			strcat(strcat(strcpy(key, email), ":"), client_addr);
			cache_insert(key, rc);
			free(key);
		}
	}
}


